@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -977,6 +977,9 @@ static int culler(fz_context *ctx, void *opaque, fz_rect bbox, fz_cull_type type
 	case FZ_CULL_PATH_FILL:
 	case FZ_CULL_PATH_STROKE:
 	case FZ_CULL_PATH_FILL_STROKE:
+	case FZ_CULL_CLIP_PATH_FILL:
+	case FZ_CULL_CLIP_PATH_STROKE:
+	case FZ_CULL_CLIP_PATH_FILL_STROKE:
 		if (red->line_art == PDF_REDACT_LINE_ART_REMOVE_IF_COVERED)
 			return (rect_touches_redactions(ctx, bbox, red) == 2);
 		else if (red->line_art == PDF_REDACT_LINE_ART_REMOVE_IF_TOUCHED)
